import os
import json
import shutil
from urllib.parse import urlparse
from typing import List
from pathlib import Path
from datetime import datetime
from hlsdl.cmdargs import CmdArgs
from hlsdl.models.base import BaseUri
from hlsdl.models.key import StreamKey
from hlsdl.models.segment import Segment
from hlsdl.util.texts import t_msg
from hlsdl.util.concat import Concat
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


class Stream:
    '''
    自适应流具体实现的父类
    每一条流应当具有以下基本属性：
        - 名称
        - 分段链接列表
        - 分辨率
        - 码率
        - 时长
        - 编码
        - 语言
    具有以下函数
        - 扩展分段 某些情况下需要合并两或多条流
        - 计算全部分段时长和大小
        - 显示流信息 总时长和大小
        - 从第一个分段读取流信息
        - 保存流相关信息至本地
        - 增加分段
        - 增加密钥信息
        - 合并 一般在下载完成之后
    '''
    def __init__(self, index: int, uri_item: BaseUri, save_dir: Path):
        self.index = index
        self.name = uri_item.name
        self.home_url = uri_item.home_url
        self.base_url = uri_item.base_url.rstrip('/')
        self.save_dir = save_dir / self.name
        self.segments = [] # type: List[Segment]
        self.duration = 0.0
        self.filesize = 0
        self.lang = ''
        self.bandwidth = None # type: int
        self.fps = None # type: int
        self.resolution = '' # type: str
        self.codecs = None # type: str
        self.streamkeys = [] # type: List[StreamKey]
        # 初始化默认设定流类型
        self.stream_type = '' # type: str
        self.model = ''
        self.suffix = '.mp4'

    def segments_extend(self, segments: List[Segment], has_init: bool = False, name_from_url: bool = False):
        '''
        某些情况下对流进行合并
        需要更新一下新增分段的文件名
        '''
        offset = len(self.segments)
        _segments = []
        for segment in segments:
            # 跳过init分段
            if segment.index == -1:
                continue
            segment.add_offset_for_name(offset, has_init, name_from_url=name_from_url)
            _segments.append(segment)
        self.segments.extend(_segments)

    def live_segments_extend(self, segments: List[Segment], has_init: bool, name_from_url: bool = False, compare_with_url: bool = False):
        '''
        对live流进行合并
        - 更新新增分段的文件名
        - 根据链接中的path部分检查是不是重复了
        '''
        if compare_with_url:
            url_paths = [segment.url for segment in self.segments if segment.skip_concat is False]
        else:
            url_paths = [urlparse(segment.url).path for segment in self.segments if segment.skip_concat is False]
        offset = len(self.segments)
        _segments = []
        for segment in segments:
            # 这里会过滤掉init分段
            if segment.index == -1:
                continue
            if compare_with_url:
                _url = segment.url
            else:
                _url = urlparse(segment.url).path
            if _url in url_paths:
                continue
            segment.set_offset_for_name(offset, has_init, name_from_url=name_from_url)
            offset += 1
            _segments.append(segment)
        self.segments.extend(_segments)

    def calc(self):
        self.duration = sum([segment.duration for segment in self.segments if segment.skip_concat is False])
        self.filesize = sum([segment.filesize for segment in self.segments if segment.skip_concat is False])
        self.filesize = self.filesize / 1024 / 1024

    def get_name(self):
        return self.name

    def get_stream_model(self):
        assert self.model != '', 'report this content to me'
        return self.model

    def check_record_time(self, live_duration: float):
        # 修正calc计算后 直接比较当前流的 duration 即可
        if live_duration == 0.0:
            return False
        return self.duration >= live_duration

    def get_init_msg(self, show_init: bool = False):
        if show_init is False:
            return ''
        for _ in self.segments:
            if _.segment_type == 'init':
                return ' initialization -> ' + _.url.split('?')[0].split('/')[-1]
        return ''

    def fix_name(self, index: int, index_to_name: bool = False):
        if index_to_name:
            self.name = f'{index}_{self.name}'

    def show_info(self, index: int, show_init: bool = False, index_to_name: bool = False):
        ''' 显示信息 '''
        self.calc()
        self.fix_name(index, index_to_name)
        if self.filesize > 0:
            print(
                f'{index:>3} {t_msg.total_segments_info_1} {len(self.segments):>4} {t_msg.total_segments_info_2} '
                f'{self.duration:>7.2f}s {self.filesize:.2f}MiB {self.get_name()}{self.get_init_msg(show_init)}'
            )
        else:
            print(
                f'{index:>3} {t_msg.total_segments_info_1} {len(self.segments):>4} {t_msg.total_segments_info_2} '
                f'{self.duration:>7.2f}s {self.get_name()}{self.get_init_msg(show_init)}'
            )

    def read_stream_header(self):
        ''' 读取一部分数据 获取流的信息 '''
        pass

    def show_segments(self):
        for segment in self.segments:
            print(segment.url)

    def dump_segments(self):
        self.calc()
        ''' 保存分段信息 '''
        # 这里的 self.get_name() 好像有种情况下会调用当前类的方法 而不是重写后的方法
        self.save_dir = self.save_dir.parent / self.get_name()
        if self.save_dir.exists() is False:
            self.save_dir.mkdir()
        keys = []
        if len(self.streamkeys) > 0:
            for streamkey in self.streamkeys:
                keys.append(streamkey.dump())
        info = {
            'name': self.get_name(),
            'path': self.save_dir.resolve().as_posix(),
            'creatTime': f'{datetime.now()}',
            'key': keys,
            'segments': [],
        }
        for segment in self.segments:
            segment.folder = self.save_dir
            info['segments'].append(
                {
                    'url': segment.url,
                    'size': segment.filesize,
                    'byterange': segment.byterange,
                    'name': segment.name,
                }
            )
        data = json.dumps(info, ensure_ascii=False, indent=4)
        (self.save_dir / 'raw.json').write_text(data, encoding='utf-8')

    def append_segment(self):
        ''' 新增分段 '''
        pass

    def append_key(self, streamkey: StreamKey):
        ''' 新增key '''
        self.streamkeys.append(streamkey)

    def fix_url(self, url: str) -> str:
        if url.startswith('http://') or url.startswith('https://') or url.startswith('ftp://'):
            return url
        elif url.startswith('/'):
            return f'{self.home_url}{url}'
        elif url == '':
            return f'{self.base_url}'
        else:
            return f'{self.base_url}/{url}'

    def concat(self, args: CmdArgs):
        ''' 合并视频 '''
        out = Path(self.save_dir.absolute().as_posix() + self.suffix)
        if args.overwrite is False and out.exists() is True:
            logger.info(f'{t_msg.try_to_concat} {self.get_name()} {t_msg.cancel_concat_reason_1}')
            return True
        skip_count = 0
        names = []
        for segment in self.segments:
            if segment.skip_concat:
                skip_count += 1
                continue
            segment_path = self.save_dir / segment.name
            if segment_path.exists() is False:
                continue
            names.append(segment.name)
        # 仅在非直播时这样比较
        if args.live is False and len(names) != len(self.segments) - skip_count:
            logger.error(f'{t_msg.try_to_concat} {self.get_name()} {t_msg.cancel_concat_reason_2}')
            return False
        if hasattr(self, "xkey") and self.xkey is not None and self.xkey.method.upper() in ['SAMPLE-AES', 'SAMPLE-AES-CTR']:
            logger.warning(t_msg.force_use_raw_concat_for_sample_aes)
            args.raw_concat = True
        if len(self.streamkeys) > 0:
            args.raw_concat = True
        ori_path = os.getcwd()
        # 需要在切换目录前获取
        os.chdir(self.save_dir.absolute().as_posix())
        cmds, _outs = Concat.gen_cmds_outs(out, names, args)
        if len(cmds) > 0 and Path(args.ffmpeg).exists() is False:
            logger.warning('ffmpeg is not exists, please put ffmpeg to binaries folder')
        for cmd in cmds:
            os.system(cmd)
        # 执行完合并命令后即刻返回原目录
        os.chdir(ori_path)
        # 合并成功则根据设定删除临时文件
        if out.exists():
            logger.info(f'{out.as_posix()} was merged successfully')
            if args.enable_auto_delete:
                shutil.rmtree(self.save_dir.absolute().as_posix())
                logger.info(f'{self.save_dir.absolute().as_posix()} was deleted')
        else:
            logger.warning(f'merge {out.as_posix()} failed')
        # 针对DASH流 如果有key 那么就解密 注意 HLS是边下边解密
        # 加密文件合并输出和临时文件夹同一级 所以前面的删除动作并不影响进一步解密
        if args.key is not None:
            if Path(args.mp4decrypt).exists() is False:
                logger.warning('mp4decrypt is not exists, please put mp4decrypt to binaries folder')
            Concat.call_mp4decrypt(out, args)
        if args.enable_auto_delete and self.save_dir.exists():
            shutil.rmtree(self.save_dir.absolute().as_posix())
        return True

    def fix_header(self, is_fake: bool):
        pass


class StreamKey:
    '''
    一条流的加密信息 含有下列属性
    - 加密方法
    - key的uri
    - key内容
    - keyid
    - 偏移量
    - 其他属性 -> 根据流类型而定
    含有下面的方法
    - 设置key内容
    - 设置iv
    - 保存为文本
    - 从文本加载
    '''
    def __init__(self):
        self.method = 'AES-128' # type: str
        self.uri = None # type: str
        self.key = b'' # type: bytes
        self.keyid = None # type: str
        self.iv = '0' * 32 # type: str

    def set_key(self, key: bytes):
        self.key = key
        return self

    def set_iv(self, iv: str):
        if iv is None:
            return
        self.iv = iv
        return self

    def dump(self):
        pass

    def load(self):
        pass


from ..mpditem import MPDItem


class S(MPDItem):
    '''
    5.3.9.6 Segment timeline
    - t -> presentationTimeOffset
    - d -> duration
    - r -> repeat
    '''
    def __init__(self, name: str):
        super(S, self).__init__(name)
        self.t = None # type: int
        self.d = None # type: int
        self.r = None # type: int

    def generate(self):
        if self.t is None:
            self.t = 0
        if self.d is None:
            self.d = 0
        if self.r is None:
            self.r = 0
        self.to_int()
        if self.r != -1:
            self.r += 1

    def to_int(self):
        self.t = int(self.t)
        self.d = int(self.d)
        self.r = int(self.r)


import base64
from typing import List
from pathlib import Path
from hlsdl.cmdargs import CmdArgs
from hlsdl.models.base import BaseUri
from hlsdl.models.stream import Stream
from hlsdl.extractors.hls.segment import HLSSegment
from hlsdl.extractors.hls.ext.xkey import XKey
from hlsdl.extractors.hls.ext.xmedia import XMedia
from hlsdl.extractors.hls.ext.xdaterange import XDateRange
from hlsdl.extractors.hls.ext.xstream_inf import XStreamInf
from hlsdl.extractors.hls.ext.xprogram_date_time import XProgramDateTime
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


class HLSStream(Stream):
    '''
    自适应流的具体实现，HLS/DASH等，为了下载对应的流，
    每一条流应当具有以下基本属性：
    - 名称
    - 分段链接列表
    - 分辨率
    - 码率
    - 时长
    - 编码
    一些可选的属性
    - 语言
    '''
    def __init__(self, index: int, uri_item: BaseUri, save_dir: Path, parent_stream: 'HLSStream'):
        super(HLSStream, self).__init__(index, uri_item, save_dir)
        self.model = 'hls'
        self.segments = [] # type: List[HLSSegment]
        # <------对于HLS类型的流额外的属性------>
        self.xkey = None
        self.program_id = None # type: int
        self.average_bandwidth = None # type: int
        self.size = None # type: int
        self.quality = None # type: str
        self.has_map_segment = False
        self.xmedias = [] # type: List[XMedia]
        self.xstream_inf = None # type: XStreamInf
        self.bakcup_xkey = None
        self.group_id = ''
        # 如果parent_stream不为空 那么将一些属性进行赋值
        if parent_stream is not None:
            self.fps = parent_stream.fps
            self.lang = parent_stream.lang
            self.group_id = parent_stream.group_id
            self.codecs = parent_stream.codecs
            self.bandwidth = parent_stream.bandwidth
            self.resolution = parent_stream.resolution
            self.stream_type = parent_stream.stream_type
        # 初始化默认设定一个分段
        self.append_segment()

    def set_stream_lang(self, lang: str):
        if not lang:
            return
        self.lang = lang

    def set_stream_group_id(self, group_id: str):
        if not group_id:
            return
        self.group_id = group_id

    def set_stream_type(self, stream_type: str):
        ''' #EXT-X-MEDIA 会表明流类型 '''
        self.stream_type = stream_type
        if self.stream_type == 'audio':
            self.suffix = '.m4a'

    def get_name(self):
        if self.stream_type != '':
            base_name = f'{self.name}_{self.stream_type}'
        else:
            base_name = self.name
        if self.resolution != '':
            base_name += f'_{self.resolution}'
        if self.lang != '':
            base_name += f'_{self.lang}'
        if self.group_id != '':
            base_name += f'_{self.group_id}'
        if self.bandwidth is not None:
            base_name += f'_{self.bandwidth / 1000:.2f}kbps'
        return base_name

    def patch_stream_info(self, stream: 'HLSStream'):
        if stream.codecs is not None:
            self.codecs = stream.codecs
        if stream.resolution != '':
            self.resolution = stream.resolution
        if stream.lang != '':
            self.lang = stream.lang
        if stream.fps is not None:
            self.fps = stream.fps
        if stream.bandwidth is not None:
            self.bandwidth = stream.bandwidth

    def set_name(self, name: str):
        self.name = name
        return self

    def set_tag(self, tag: str):
        self.tag = tag

    def set_map_flag(self):
        self.has_map_segment = True
        self.name += f'_{self.index}'
        self.save_dir = (self.save_dir.parent / self.name).resolve()
        for segment in self.segments:
            segment.set_folder(self.save_dir)

    def append_segment(self):
        '''
        新增一个分段
        '''
        index = len(self.segments)
        if self.has_map_segment:
            index -= 1
        segment = HLSSegment().set_index(index).set_folder(self.save_dir)
        self.segments.append(segment)

    def try_fetch_key(self, args: CmdArgs):
        '''
        在解析过程中 已经设置了key的信息了
        但是没有请求key 这里是独立加载key的部分
        放在这个位置的原因是
            - 解析过程其实很短，没必要在解析时操作
            - 解析后还有合并流的过程
        所以最佳的方案是在解析之后再进行key的加载
        '''
        custom_xkey = XKey()
        if args.b64key is not None:
            _key = base64.b64decode(args.b64key.encode('utf-8'))
            custom_xkey.set_key(_key)
            if args.hexiv is not None:
                custom_xkey.set_iv(args.hexiv)
        if self.xkey is None:
            if args.b64key:
                # 如果解析后没有密钥相关信息
                # 而命令行又指定了 也进行设定
                self.set_segments_key(custom_xkey)
            return
        if self.xkey.method and self.xkey.method.upper() in ['SAMPLE-AES', 'SAMPLE-AES-CTR']:
            return
        if self.xkey.load(args, custom_xkey) is True:
            logger.info(f'm3u8 key loaded\nmethod => {self.xkey.method}\nkey    => {self.xkey.key}\niv     => {self.xkey.iv}')
            self.set_segments_key(self.xkey)

    def set_segments_key(self, xkey: XKey):
        '''
        和每个分段的key对比 设定对应的解密信息
        '''
        self.xkey = xkey
        for segment in self.segments:
            segment.xkey = xkey

    def set_xkey(self, xkey: XKey):
        self.xkey = xkey

    def set_bakcup_xkey(self, bakcup_xkey: XKey):
        self.bakcup_xkey = bakcup_xkey

    def set_xstream_inf(self, line: str):
        self.xstream_inf = XStreamInf().set_attrs_from_line(line)
        if self.xstream_inf is not None:
            self.fps = self.xstream_inf.fps
            self.codecs = self.xstream_inf.codecs
            self.bandwidth = self.xstream_inf.bandwidth
            self.resolution = self.xstream_inf.resolution
            self.stream_type = self.xstream_inf.streamtype

    def get_path(self):
        ''' 某些m3u8会有重复的 例如D+ 这里辅助去重 '''
        # return self.origin_url.split('?', maxsplit=1)[0].split('/')[-1]
        return self.origin_url.split('?', maxsplit=1)[0].split('/', maxsplit=3)[-1:]

    def set_url(self, home_url: str, base_url: str, line: str):
        if line.startswith('http://') or line.startswith('https://') or line.startswith('ftp://'):
            self.origin_url = line
        elif line.startswith('/'):
            self.origin_url = f'{home_url}{line}'
        else:
            self.origin_url = f'{base_url}/{line}'

    def set_key(self, home_url: str, base_url: str, line: str):
        xkey = XKey().set_attrs_from_line(home_url, base_url, line)
        if xkey is None:
            return
        if self.xkey is None:
            self.xkey = xkey
        else:
            self.bakcup_xkey = xkey

    def set_media(self, home_url: str, base_url: str, line: str):
        xmedia = XMedia()
        xmedia.set_attrs_from_line(line)
        xmedia.uri = self.set_origin_url(home_url, base_url, xmedia.uri)
        self.set_stream_lang(xmedia.language)
        self.set_stream_group_id(xmedia.group_id)
        self.set_stream_type(xmedia.type)
        self.xmedias.append(xmedia)

    def set_origin_url(self, home_url: str, base_url: str, uri: str):
        # 某些标签 应该被视作一个新的Stream 所以要设置其对应的原始链接
        if uri.startswith('http://') or uri.startswith('https://') or uri.startswith('ftp://'):
            self.origin_url = uri
        elif uri.startswith('/'):
            self.origin_url = f'{home_url}{uri}'
            # 更新 base_url
            self.base_url = self.origin_url.split('?', maxsplit=1)[0][::-1].split('/', maxsplit=1)[-1][::-1]
        else:
            self.origin_url = f'{base_url}/{uri}'
            # 更新 base_url
            self.base_url = self.origin_url.split('?', maxsplit=1)[0][::-1].split('/', maxsplit=1)[-1][::-1]
        return self.origin_url

    def set_daterange(self, line: str):
        self.xdaterange = XDateRange().set_attrs_from_line(line)

    def set_xprogram_date_time(self, line: str):
        self.xprogram_date_time = XProgramDateTime().set_attrs_from_line(line)


from ..ismitem import ISMItem


class StreamIndex(ISMItem):
    def __init__(self, name: str):
        super(StreamIndex, self).__init__(name)
        # <----- 通用 ----->
        self.Type = None # type: str
        self.QualityLevels = None # type: int
        self.TimeScale = None # type: int
        self.Name = None # type: str
        self.Chunks = None # type: int
        self.Url = None # type: str
        # <----- 视频 ----->
        self.MaxWidth = None # type: str
        self.MaxHeight = None # type: str
        self.DisplayWidth = None # type: str
        self.DisplayHeight = None # type: str
        # <----- 字幕 & 音频 ----->
        self.Language = None # type: str
        # <----- 字幕 ----->
        self.Subtype = None # type: str
        self.Language = None # type: str

    def generate(self):
        self.to_int('QualityLevels')
        self.to_int('Chunks')
        self.to_int('TimeScale')

    def get_media_url(self) -> str:
        return self.Url


from typing import List
from .ism import ISM
from .childs.c import c as Cc
from .childs.protectionheader import ProtectionHeader
from .childs.streamindex import StreamIndex
from .childs.qualitylevel import QualityLevel
from .handler import xml_handler

from hlsdl.cmdargs import CmdArgs
from hlsdl.models.base import BaseUri
from hlsdl.extractors.mss.key import MSSKey
from hlsdl.extractors.mss.stream import MSSStream
from hlsdl.extractors.base import BaseParser
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


class MSSParser(BaseParser):
    def __init__(self, args: CmdArgs, uri_type: str):
        super(MSSParser, self).__init__(args, uri_type)
        self.suffix = '.ism'

    def parse(self, uri: str, content: str) -> List[MSSStream]:
        uri_item = self.parse_uri(uri)
        assert uri_item is not None, f'parse {uri} failed'
        self.dump_content(uri_item.name, content, self.suffix)
        # 解析转换内容为期望的对象
        ism = xml_handler(content)
        return self.walk_streamindex(ism, uri_item)

    def walk_streamindex(self, ism: ISM, uri_item: BaseUri) -> List[MSSStream]:
        streamindexs = ism.find('StreamIndex') # type: List[StreamIndex]
        assert len(streamindexs) > 0, 'there is no StreamIndex'
        # 遍历处理streamindexs
        streams = [] # type: List[MSSStream]
        for streamindex in streamindexs:
            streams.extend(self.walk_qualitylevel(streamindex, ism, len(streams), uri_item))
        # 处理空分段
        for stream in streams:
            if stream.segments[-1].url == '':
                _ = stream.segments.pop(-1)
        return streams

    def walk_qualitylevel(self, streamindex: StreamIndex, ism: ISM, sindex: int, uri_item: BaseUri) -> List[MSSStream]:
        streams = [] # type: List[MSSStream]
        qualitylevels = streamindex.find('QualityLevel') # type: List[QualityLevel]
        if len(qualitylevels) == 0:
            return streams
        for track_index, qualitylevel in enumerate(qualitylevels):
            stream = MSSStream(sindex + len(streams), uri_item, self.args.save_dir)
            stream.set_track_index(track_index)
            streams.extend(self.walk_c(qualitylevel, streamindex, ism, stream))
        return streams

    def walk_c(self, qualitylevel: QualityLevel, streamindex: StreamIndex, ism: ISM, stream: MSSStream) -> List[MSSStream]:
        cs = streamindex.find('c') # type: List[Cc]
        if len(cs) == 0:
            return []
        # 设置基本信息
        stream.set_track_name(streamindex.Name)
        stream.set_timescale(streamindex.TimeScale or ism.TimeScale)
        stream.set_stream_type(streamindex.Type)
        stream.set_codecs(qualitylevel.FourCC)
        stream.set_bits_per_sample(qualitylevel.BitsPerSample)
        stream.set_sampling_rate(qualitylevel.SamplingRate)
        stream.set_channels(qualitylevel.Channels)
        stream.set_codec_private_data(qualitylevel.CodecPrivateData)
        stream.set_nal_unit_length_field(qualitylevel.NALUnitLengthField)
        stream.set_bandwidth(qualitylevel.Bitrate)
        stream.set_lang(streamindex.Language)
        stream.set_resolution(qualitylevel.MaxWidth, qualitylevel.MaxHeight)
        last_end_time = None # type: int
        for c in cs:
            for _ in range(c.r):
                media_url = streamindex.get_media_url()
                if last_end_time is None:
                    if c.t is None:
                        last_end_time = 0
                    else:
                        last_end_time = c.t
                if '{bitrate}' in media_url:
                    media_url = media_url.replace('{bitrate}', str(qualitylevel.Bitrate))
                if '{start time}' in media_url:
                    media_url = media_url.replace('{start time}', str(last_end_time))
                if streamindex.TimeScale is not None:
                    duration = c.d / streamindex.TimeScale
                elif ism.TimeScale is not None:
                    duration = c.d / ism.TimeScale
                else:
                    duration = c.d / 10000000
                stream.set_segment_duration(duration)
                stream.set_media_url(media_url)
                last_end_time += c.d
        protections = ism.find('Protection')
        if len(protections) > 0 and len(protections[0].find('ProtectionHeader')) > 0:
            protection_flag = True
            protectionheaders = protections[0].find('ProtectionHeader') # type: List[ProtectionHeader]
            if len(protectionheaders) > 1:
                print('please report this ism content to me')
            for protectionheader in protectionheaders:
                stream.set_kid(protectionheader.kid)
                stream.append_key(MSSKey(protectionheader))
            logger.debug(f'ProtectionHeader was found')
        else:
            protection_flag = False
        stream.set_protection_flag(protection_flag)
        return [stream]


from hlsdl.models.key import StreamKey
from .childs.protectionheader import ProtectionHeader


class MSSKey(StreamKey):
    def __init__(self, ph: ProtectionHeader):
        super(MSSKey, self).__init__()
        self.systemid = ph.SystemID
        self.key = ph.innertext


from .x import X


class XStreamInf(X):
    '''
    #EXT-X-STREAM-INF 紧接着该标签的实际上是一个Stream
    - PROGRAM-ID=1,BANDWIDTH=1470188,SIZE=468254984,FPS=25,RESOLU=1080,CODECS="avc1,mp4a",QUALITY=5,STREAMTYPE="mp4hd3"
    #EXT-X-I-FRAME-STREAM-INF 也被归类于此 有一个额外属性 URI
    '''
    def __init__(self):
        super(XStreamInf, self).__init__('#EXT-X-STREAM-INF')
        self.program_id = None # type: int
        self.bandwidth = None # type: int
        self.average_bandwidth = None # type: int
        self.codecs = None # type: str
        self.resolution = '' # type: str
        self.frame_rate = None # type: float
        self.hdcp_level = None # type: str
        self.characteristics = None # type: str
        self.uri = None # type: str
        self.audio = None # type: str
        self.video = None # type: str
        self.subtitles = None # type: str
        self.closed_captions = None # type: str
        self.video_range = None # type: str
        self.size = None # type: int
        self.fps = None # type: float
        self.quality = None # type: int
        self.streamtype = '' # type: str
        # VIDEO-RANGE是苹果的标准 往下的是非标准属性
        self.known_attrs = {
            'PROGRAM-ID': int,
            'BANDWIDTH': int,
            'AVERAGE-BANDWIDTH': int,
            'CODECS': 'codecs',
            'RESOLUTION': 'resolution',
            'FRAME-RATE': float,
            'HDCP-LEVEL': 'hdcp_level',
            'CHARACTERISTICS': 'characteristics',
            'URI': 'uri',
            'AUDIO': 'audio',
            'VIDEO': 'video',
            'SUBTITLES': 'subtitles',
            'CLOSED-CAPTIONS': 'closed_captions',
            'VIDEO-RANGE': 'video_range',
            'SIZE': int,
            'FPS': float,
            'RESOLU': 'resolution',
            'QUALITY': int,
            'STREAMTYPE': 'streamtype',
        }

    def set_attrs_from_line(self, line: str):
        '''
        这里实际上可以不写
        '''
        return super(XStreamInf, self).set_attrs_from_line(line)


import asyncio
import platform
from typing import List
from pathlib import Path
from aiohttp.connector import TCPConnector
from aiohttp import ClientSession, ClientResponse
from aiohttp_socks import ProxyConnector
from hlsdl.cmdargs import CmdArgs
from hlsdl.models.stream import Stream
from hlsdl.extractors.hls.parser import HLSParser
from hlsdl.extractors.hls.stream import HLSStream
from hlsdl.extractors.dash.parser import DASHParser
from hlsdl.extractors.dash.stream import DASHStream
from hlsdl.extractors.mss.parser import MSSParser
from hlsdl.extractors.mss.stream import MSSStream
from hlsdl.util.texts import t_msg

from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')


class Extractor:
    '''
    请求DASH/HLS链接内容，也就是下载元数据
    或者读取含有元数据的文件
    或者读取含有多个元数据文件的文件夹
    最终得到一个Stream（流）对象供Downloader（下载器）下载
    '''
    def __init__(self, args: CmdArgs):
        self.args = args
        self.parser = None

    def load_raw2text(self, data: bytes):
        raw_text = None # type: str
        try:
            raw_text = data.decode('utf-8')
        except UnicodeDecodeError:
            try:
                raw_text = data.decode('utf-16')
            except Exception as e:
                logger.error(f'load_raw2text failed', exc_info=e)
        return raw_text

    def fetch_metadata(self, uri: str, parent_stream: Stream = None):
        '''
        从链接/文件/文件夹等加载内容 解析metadata
        '''
        if uri.startswith('http://') or uri.startswith('https://') or uri.startswith('ftp://'):
            if platform.system() == 'Windows':
                asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
            loop = asyncio.get_event_loop()
            return self.raw2streams('url', *loop.run_until_complete(self.fetch(uri)), parent_stream)
        if '\\' in uri:
            _file_name = uri.split('\\')[-1]
        elif '/' in uri:
            _file_name = uri.split('/')[-1]
        else:
            _file_name = uri
        illegal_symbols = ["\\", "/", ":", "：", "*", "?", "\"", "<", ">", "|", "\r", "\n", "\t"]
        is_illegal_path = True
        for illegal_symbol in illegal_symbols:
            if illegal_symbol in _file_name:
                is_illegal_path = False
                break
        if is_illegal_path is False:
            return
        if Path(uri).exists() is False:
            return
        if Path(uri).is_file():
            return self.raw2streams('path', uri, self.load_raw2text(Path(uri).read_bytes()), parent_stream)
        if Path(uri).is_dir() is False:
            return
        streams = []
        for path in Path(uri).iterdir():
            _streams = self.raw2streams('path', path.name, self.load_raw2text(path.read_bytes()), parent_stream)
            if _streams is None:
                continue
            streams.extend(_streams)
        return streams

    async def fetch(self, url: str) -> str:
        if self.args.proxy != '':
            connector = ProxyConnector.from_url(self.args.proxy, ssl=False)
        else:
            connector = TCPConnector(ssl=False)
        async with ClientSession(connector=connector) as client: # type: ClientSession
            async with client.get(url, headers=self.args.headers) as resp: # type: ClientResponse
                return str(resp.url), self.load_raw2text(await resp.read())

    def raw2streams(self, uri_type: str, uri: str, content: str, parent_stream: Stream) -> List[Stream]:
        '''
        解析解码后的返回结果
        '''
        if not content:
            return []
        if content.startswith('#EXTM3U'):
            return self.parse_as_hls(uri_type, uri, content, parent_stream)
        elif '<MPD' in content and '</MPD>' in content:
            return self.parse_as_dash(uri_type, uri, content, parent_stream)
        elif '<SmoothStreamingMedia' in content and '</SmoothStreamingMedia>' in content:
            return self.parse_as_mss(uri_type, uri, content, parent_stream)
        else:
            logger.warning(t_msg.cannot_get_stream_metadata)
            return []

    def parse_as_hls(self, uri_type: str, uri: str, content: str, parent_stream: HLSStream = None) -> List[HLSStream]:
        _streams = HLSParser(self.args, uri_type).parse(uri, content, parent_stream)
        streams = []
        for stream in _streams:
            # 针对master类型加载详细内容
            if stream.tag != '#EXT-X-STREAM-INF' and stream.tag != '#EXT-X-MEDIA':
                streams.append(stream)
                continue
            if self.args.hide_load_metadata:
                logger.debug(f'Load {stream.tag} metadata from -> {stream.origin_url}')
            else:
                logger.info(f'Load {stream.tag} metadata from -> {stream.origin_url}')
            new_streams = self.fetch_metadata(stream.origin_url, parent_stream=stream)
            if new_streams is None:
                continue
            if len(new_streams) == 1:
                new_streams[0].patch_stream_info(stream)
            streams.extend(new_streams)
        # 在全部流解析完成后 再处理key
        for stream in streams:
            stream.try_fetch_key(self.args)
        return streams

    def parse_as_dash(self, uri_type: str, uri: str, content: str, parent_stream: DASHStream = None) -> List[DASHStream]:
        self.parser = DASHParser(self.args, uri_type)
        streams = self.parser.parse(uri, content)
        return streams

    def parse_as_mss(self, uri_type: str, uri: str, content: str, parent_stream: MSSStream = None) -> List[MSSStream]:
        streams = MSSParser(self.args, uri_type).parse(uri, content)
        return streams


import re


class X:
    '''
    每一个标签具有的通用性质
    - 标签名
    - 以期望的形式打印本身信息
    - 标签行去除 TAG_NAME: 部分
    '''
    def __init__(self, TAG_NAME: str = 'X'):
        self.TAG_NAME = TAG_NAME
        self.known_attrs = {}

    def __repr__(self):
        return f'{self.TAG_NAME}'

    def __strip(self, line: str):
        # return line[len(self.TAG_NAME) + 1:]
        data = line.split(':', maxsplit=1)
        if len(data) == 2:
            _, no_tag_line = data
        elif len(data) == 0:
            raise 'm3u8格式错误 无法处理的异常'
        else:
            no_tag_line = data[0]
        return no_tag_line

    def get_tag_info(self, line: str):
        return self.__strip(line)

    def format_key(self, key: str):
        return key.replace('-', '_').lower()

    def convert_type(self, name: str, value: str, _type: type):
        self.__setattr__(self.format_key(name), _type(value))

    def regex_attrs(self, info: str) -> list:
        if info.endswith(',') is False:
            info += ','
        return re.findall('(.*?)=("[^"]*?"|[^,]*?),', info)

    def set_attrs_from_line(self, line: str):
        '''
        https://stackoverflow.com/questions/34081567
        re.findall('([A-Z]+[0-9]*)=("[^"]*"|[^,]*)', s)
        '''
        info = self.get_tag_info(line)
        for key, value in self.regex_attrs(info):
            value = value.strip('"')
            if key in self.known_attrs:
                if isinstance(self.known_attrs[key], str):
                    self.__setattr__(self.known_attrs[key], value)
                elif isinstance(self.known_attrs[key], type):
                    self.convert_type(key, value, self.known_attrs[key])
                else:
                    self.known_attrs[key](value)
            else:
                print(f'unknown attr -> {key} <- of {self.TAG_NAME}')
        return self


import asyncio
import platform
from aiohttp_socks import ProxyConnector
from aiohttp import ClientSession, ClientResponse
from aiohttp.connector import TCPConnector

from .x import X
from hlsdl.cmdargs import CmdArgs
from hlsdl.log import setup_logger

logger = setup_logger('XstreamDL', level='INFO')

DEFAULT_IV = '0' * 32


class XKey(X):
    '''
    一组加密参数
    - METHOD
        - AES-128
        - SAMPLE-AES
    - URI
        - data:text/plain;base64,...
        - skd://...
    - IV
        - 0x/0X...
    - KEYFORMAT
        - urn:uuid:edef8ba9-79d6-4ace-a3c8-27dcd51d21ed
        - com.apple.streamingkeydelivery
    '''
    def __init__(self):
        super(XKey, self).__init__('#EXT-X-KEY')
        self.method = 'AES-128' # type: str
        self.uri = None # type: str
        self.key = b'' # type: bytes
        self.keyid = None # type: str
        self.iv = DEFAULT_IV # type: str
        self.keyformatversions = None # type: str
        self.keyformat = None # type: str
        self.known_attrs = {
            'METHOD': 'method',
            'URI': 'uri',
            'KEYID': 'keyid',
            'IV': 'iv',
            'KEYFORMATVERSIONS': 'keyformatversions',
            'KEYFORMAT': 'keyformat',
        }

    def set_key(self, key: bytes):
        self.key = key
        return self

    def set_iv(self, iv: str):
        if iv is None:
            return
        self.iv = iv
        return self

    def set_attrs_from_line(self, home_url: str, base_url: str, line: str):
        '''
        key的链接可能不全 用home_url或base_url进行补齐 具体处理后面做
        '''
        line = line.replace('MEATHOD', 'METHOD')
        super(XKey, self).set_attrs_from_line(line)
        key_type, self.uri = self.gen_hls_key_uri(home_url, base_url)
        if self.iv.lower().startswith('0x'):
            self.iv = self.iv[2:]
        return self

    def gen_hls_key_uri(self, home_url: str, base_url: str):
        '''
        解析时 不具体调用这个函数 需要的地方再转换
        data:text/plain;base64,AAAASnBzc2gAAAAA7e+LqXnWSs6jyCfc1R0h7QAAACoSEKg079lX5xeK9g/zZPwXENESEKg079lX5xeK9g/zZPwXENFI88aJmwY=
        skd://a834efd957e7178af60ff364fc1710d1
        '''
        if self.uri is None:
            return '', self.uri
        if self.uri.startswith('data:text/plain;base64,'):
            return 'base64', self.uri.split(',', maxsplit=1)[-1]
        elif self.uri.startswith('skd://'):
            return 'skd', self.uri.split('/', maxsplit=1)[-1]
        elif self.uri.startswith('http'):
            return 'http', self.uri
        elif self.uri.startswith('/'):
            return 'http', home_url + self.uri
        else:
            return 'http', base_url + '/' + self.uri

    async def fetch(self, url: str, args: CmdArgs) -> bytes:
        if args.proxy != '':
            connector = ProxyConnector.from_url(args.proxy, ssl=False)
        else:
            connector = TCPConnector(ssl=False)
        async with ClientSession(connector=connector) as client: # type: ClientSession
            async with client.get(url, headers=args.headers) as resp: # type: ClientResponse
                return await resp.content.read()

    def load(self, args: CmdArgs, custom_xkey: 'XKey'):
        '''
        如果custom_xkey存在key 那么覆盖解析结果中的key
        并且不进行请求key的动作 同时覆盖iv 如果有自定义iv的话
        '''
        if custom_xkey.iv != DEFAULT_IV:
            self.iv = custom_xkey.iv
        if custom_xkey.key != b'':
            self.key = custom_xkey.key
            return True
        if self.uri.startswith('http://') or self.uri.startswith('https://'):
            logger.info(f'key uri => {self.uri}')
            if platform.system() == 'Windows':
                asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
            loop = asyncio.get_event_loop()
            self.key = loop.run_until_complete(self.fetch(self.uri, args))
        elif self.uri.startswith('ftp://'):
            return False
        return True